#include <polylib/polylibgmp.h>
#include "combine.h"

/*
 * Vector p3 is a linear combination (a p1 + b p2) of two vectors
 * (p1 and p2) such that a >= 0 and p3[pos] is zero.
 * Since b may be negative, p2 is assumed to be an equality.
 * In particular, if g = gcd(p1[pos], p2[pos]), then a = |p2[pos]|/g and
 * b = -sgn(p2[pos]) p1[pos]/g.
 * The first element of each vector (p1,p2,p3) is a status
 * element and is not changed in p3.
 * The parameter 'length' does not include status element one.
 * 
 * This function was inspired by PolyLib.
 */
void Combine(Value *p1, Value *p2, Value *p3, int pos, unsigned length)
{
    Value a;
    Value b;

    assert(value_zero_p(p2[0]));

    value_init(a);
    value_init(b);

    value_gcd(b, p1[pos], p2[pos]);
    value_division(a, p2[pos], b);
    value_absolute(a, a);
    value_division(b, p1[pos], b);
    if (value_pos_p(p2[pos]))
	value_oppose(b, b);

    Vector_Combine(p1+1, p2+1, p3+1, a, b, length);
    Vector_Normalize(p3+1, length);

    value_clear(a);
    value_clear(b);
}
