#ifndef DOMAIN_H
#define DOMAIN_H

#include "suif_inc.h"
#include "srel.h"

#include <map>
#include <vector>

using std::map;
//using std::vector;

extern const char * k_static_root;
extern const char * k_iteration_domain;
extern const char * k_domain_symtab;

typedef std::vector<var_sym *> var_sym_v; // why is std:: needed ?

struct Param_Iterator : public Global_Var_Decl {
    Param_Iterator(Const_String name) : Global_Var_Decl(name), iterator(false) {}
    bool iterator;
};

struct symtab_ann {
    global_symtab *	symtab;
    map<var_sym *, Param_Iterator *> vars;
    void add_var(var_sym *var);
};

class guard_constructor {
    int			verbose;
    symtab_ann *	sa;
    F_And *		f;
    F_Exists *		e;

public:
    Relation *	    	r;
    guard_constructor(symtab_ann *sa, int verbose) : sa(sa), verbose(verbose) {
	r = new Relation(0);
	e = r->add_exists();
	f = e->add_and();
    }
    ~guard_constructor() {
	delete r;
    }

    boolean construct_guard(instruction *ins);
    boolean fill_q(Constraint_Handle& h, operand op, int mult);
};

struct context_info {
    symtab_ann *sa;
    SRelation  *ctx;
    tree_proc  *tp;

    context_info() : sa(NULL), ctx(NULL), tp(NULL) {}
};

struct stmt;
struct stmt_node;
struct stmt_access;

class domain_constructor {
    SRelation * 	c;
    symtab_ann *	sa;
    int			level;
    int			plevel;
    var_sym_v		its;
    var_sym_v		out;
    instruction *	root;
    int			verbose;
    int			oldstyle;
    std::vector<int>	prefix;
    tree_proc *		tp;

    SRelation * reorder(SRelation * c);
public:
    named_lin_ineq * reorder(named_lin_ineq * c, int params = -1,
						 int output = 0);
    domain_constructor(tree_proc *tp, int verbose, int oldstyle) : 
	tp(tp), verbose(verbose), oldstyle(oldstyle) {}
    void scan(context_info *ctx);
    boolean scan(tree_node_list * tp);
    boolean scan(tree_node * tp);
    void construct();
    void construct(tree_node_list * tp);
    void construct(tree_node * tp);
    void recurse(tree_node * tp);
    stmt *extract_stmt(tree_instr *i);
    stmt_node *extract_stmt_node(instruction *i);
    stmt_node *extract_stmt_node(operand &op, boolean write);
    stmt_node *extract_stmt_node_cal(in_cal *cal);
    stmt_access *create_access(dep_instr_annote *di, bool write);
    stmt_access *check_access(instruction *in, bool write);
    void visit(tree_node *t);
    boolean construct_guard(tree_if *ti, tree_node_list *tnl);
    void set_prefix_and_iterator(int prefix_value, var_sym *iterator = NULL);
};

struct stmt_node {
	virtual ~stmt_node() {}
	virtual bool is_access() = 0;
	virtual void reorder(domain_constructor *dc) {}
};

struct stmt_access : stmt_node {
	bool			 write;
	named_lin_ineq      	*nlq;
	var_sym			*array;
	int		 	 output;

	stmt_access(bool w, named_lin_ineq *nlq, var_sym *a, int o) :
		write(w), nlq(nlq), array(a), output(o) {}
	virtual bool is_access() { return true; }
	~stmt_access() {
		delete nlq;
	}
	virtual void reorder(domain_constructor *dc) {
		named_lin_ineq *t = nlq;
		nlq = dc->reorder(nlq, -1, output);
		delete t;
	}
};

struct stmt_call : stmt_node {
	const char *name;
	std::vector<stmt_node *>	arg;

	stmt_call(const char *name) : name(name) {}
	virtual bool is_access() { return false; }

	~stmt_call() {
		for (int i = 0; i < arg.size(); ++i)
			delete arg[i];
	}
	virtual void reorder(domain_constructor *dc) {
		for (int i = 0; i < arg.size(); ++i)
			arg[i]->reorder(dc);
	}
};

struct stmt {
	std::vector<stmt_access *>	out;
	stmt_node			*tree;

	stmt() {
		tree = NULL;
	}
	~stmt() {
		for (int i = 0; i < out.size(); ++i)
			delete out[i];
		delete tree;
	}
	void reorder(domain_constructor *dc) {
		for (int i = 0; i < out.size(); ++i)
			out[i]->reorder(dc);
		tree->reorder(dc);
	}
};

struct domain_ann {
    instruction	   	*insn;
    SRelation      	*domain;
    int			 level;
    int			 plevel;
    var_sym_v		 its;
    std::vector<int>	 prefix;
    stmt		*statement;

    domain_ann(instruction *i, SRelation *d, int l, int pl, var_sym_v it,
		std::vector<int> p, stmt *s):
	insn(i), domain(d), level(l), plevel(pl), its(it), prefix(p),
	statement(s) {}
};

struct access_ann {
    boolean		 write;
    named_lin_ineq      *nlq;
    var_sym		*array;
    int		 	 output;
    int			 arg_pos; /* argument position in function call or -1 */

    access_ann(boolean w, named_lin_ineq *n, var_sym *a, int o, int pos) :
	write(w), nlq(n), array(a), output(o), arg_pos(pos) {}
};

void construct_domains(tree_proc * tp, int verbose, int oldstyle, context_info *ctx);
void init(int &argc, char *argv[]);

#endif
