/* file "for_normalize.cc" */

/*

    Copyright (C) 1994 Stanford University

    All rights reserved.

    NOTICE:  This software is provided ``as is'', without any
    warranty, including any implied warranty for merchantability or
    fitness for a particular purpose.  Under no circumstances shall
    Stanford University or its agents be liable for any use of, misuse
    of, or inability to use this software, including incidental and
    consequential damages.

    License is hereby given to use, modify, and redistribute this
    software, in whole or in part, for any purpose, commercial or
    non-commercial, provided that the user agrees to the terms of this
    copyright notice, including disclaimer of warranty, and provided
    that this copyright notice, including disclaimer of warranty, is
    preserved in the source code and documentation of anything derived
    from this software.  Any redistributor of this software or
    anything derived from this software assumes responsibility for
    ensuring that any parties to whom such a redistribution is made
    are fully aware of the terms of this license and disclaimer.

    "SUIF" is a trademark of Stanford University.


*/

/* code to normalize for loops to have zero lower bound and step size
 * of one for the porky program for SUIF */

#include "for_normalize.h"

static instruction_list *normalized;

static void normalize_fors_on_node(tree_node *the_node, void *);

extern instruction_list *normalized_fors_on_proc(tree_proc *the_proc)
  {
    normalized = new instruction_list;
    the_proc->map(&normalize_fors_on_node, NULL);
    if (normalized->is_empty()) {
	delete normalized;
	normalized = NULL;
    }
    return normalized;
  }

static void normalize_fors_on_node(tree_node *the_node, void *)
  {
    if (!the_node->is_for())
        return;

    tree_for *the_for = (tree_for *)the_node;

    int step_int;
    boolean step_is_int = operand_is_int_const(the_for->step_op(), &step_int);

    if (step_is_int && (step_int == 1))
        return;

    guard_for(the_for);

    if (!operand_is_int_const(the_for->lb_op()))
        make_lb_temp(the_for);
    if (!operand_is_int_const(the_for->step_op()))
        make_step_temp(the_for);

    operand old_lb_op = the_for->lb_op();
    operand old_step_op = the_for->step_op();

    if (!operand_is_int_const(the_for->ub_op()))
        make_ub_temp(the_for);

    tree_for_test new_test;
    switch (the_for->test())
      {
        case FOR_SGTE:
        case FOR_SLTE:
        case FOR_SGT:
        case FOR_SLT:
            new_test = FOR_SLTE;
            break;
        case FOR_UGTE:
        case FOR_ULTE:
        case FOR_UGT:
        case FOR_ULT:
            new_test = FOR_ULTE;
            break;
        default:
            return;
      }

    var_sym *old_index = the_for->index();
    type_node *index_type = old_index->type()->unqual();
    var_sym *new_index = the_for->scope()->new_unique_var(index_type);

    operand new_ub_op = cast_op(iteration_count(the_for) - 1, index_type);

    old_lb_op.remove();
    old_step_op.remove();

    operand old_ub_op = the_for->ub_op();
    old_ub_op.remove();
    if (old_ub_op.is_expr())
        delete old_ub_op.instr();

    the_for->set_lb_op(operand_int(index_type, 0));
    the_for->set_ub_op(new_ub_op);
    the_for->set_step_op(operand_int(index_type, 1));

    the_for->set_test(new_test);

    the_for->set_index(new_index);

    operand old_ind_op =
#ifndef DEAL_WITH_SOLARIS_BRAIN_DAMAGE
            cast_op((new_index * old_step_op) + old_lb_op, index_type);
#else
            cast_op(fold_mul(new_index, old_step_op) + old_lb_op, index_type);
#endif
    tree_node *assign_node = assign(old_index, old_ind_op);
    tree_node *assign_node2 = assign_node->clone();
    the_for->body()->push(assign_node2);
    the_for->parent()->insert_after(assign_node, the_for->list_e());
    normalized->push(((tree_instr *)assign_node)->instr());
    normalized->push(((tree_instr *)assign_node2)->instr());
  }

/*----------------------------------------------------------------------*
    End Private Function Implementations
 *----------------------------------------------------------------------*/
